from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from django.apps import AppConfig

import scout_apm.core
from scout_apm.django.config import ConfigAdapter
from scout_apm.django.instruments.sql import SQLInstrument
from scout_apm.django.instruments.template import TemplateInstrument

logger = logging.getLogger(__name__)


class ScoutApmDjangoConfig(AppConfig):
    name = "scout_apm"
    verbose_name = "Scout Apm (Django)"

    def ready(self):
        # Copy django configuration to scout_apm's config
        ConfigAdapter.install()

        # Finish installing the agent. If the agent isn't installed for any
        # reason, return without installing instruments
        installed = scout_apm.core.install()
        if not installed:
            return

        self.install_middleware()

        # Setup Instruments
        SQLInstrument.install()
        TemplateInstrument.install()

    def install_middleware(self):
        """
        Attempts to insert the ScoutApm middleware as the first middleware
        (first on incoming requests, last on outgoing responses).
        """
        from django.conf import settings

        # If MIDDLEWARE is set, update that, with handling of tuple vs array forms
        if getattr(settings, "MIDDLEWARE", None) is not None:
            if isinstance(settings.MIDDLEWARE, tuple):
                settings.MIDDLEWARE = (
                    ("scout_apm.django.middleware.MiddlewareTimingMiddleware",)
                    + settings.MIDDLEWARE
                    + ("scout_apm.django.middleware.ViewTimingMiddleware",)
                )
            else:
                settings.MIDDLEWARE.insert(
                    0, "scout_apm.django.middleware.MiddlewareTimingMiddleware"
                )
                settings.MIDDLEWARE.append(
                    "scout_apm.django.middleware.ViewTimingMiddleware"
                )

        # Otherwise, we're doing old style middleware, do the same thing with
        # the same handling of tuple vs array forms
        else:
            if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                settings.MIDDLEWARE_CLASSES = (
                    ("scout_apm.django.middleware.OldStyleMiddlewareTimingMiddleware",)
                    + settings.MIDDLEWARE_CLASSES
                    + ("scout_apm.django.middleware.OldStyleViewMiddleware",)
                )
            else:
                settings.MIDDLEWARE_CLASSES.insert(
                    0, "scout_apm.django.middleware.OldStyleMiddlewareTimingMiddleware"
                )
                settings.MIDDLEWARE_CLASSES.append(
                    "scout_apm.django.middleware.OldStyleViewMiddleware"
                )



from __future__ import absolute_import, division, print_function, unicode_literals

from sqlalchemy import event

from scout_apm.core.tracked_request import TrackedRequest


def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    tr = TrackedRequest.instance()
    span = tr.start_span(operation="SQL/Query")
    span.tag("db.statement", statement)


def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    tr = TrackedRequest.instance()
    span = tr.current_span()
    if span is not None:
        tr.callset.update(statement, 1, span.duration())
        if tr.callset.should_capture_backtrace(statement):
            span.capture_backtrace()
    tr.stop_span()


def instrument_sqlalchemy(engine):
    if getattr(engine, "_scout_instrumented", False):
        return
    event.listen(engine, "before_cursor_execute", before_cursor_execute)
    event.listen(engine, "after_cursor_execute", after_cursor_execute)
    engine._scout_instrumented = True



from __future__ import absolute_import, division, print_function, unicode_literals

import scout_apm.core
from scout_apm.api.context import Context
from scout_apm.core.config import ScoutConfig
from scout_apm.core.ignore import ignore_path
from scout_apm.core.tracked_request import TrackedRequest


def includeme(config):
    configs = {}
    pyramid_config = config.get_settings()
    for name in pyramid_config:
        if name.startswith("SCOUT_"):
            value = pyramid_config[name]
            clean_name = name.replace("SCOUT_", "").lower()
            configs[clean_name] = value
    ScoutConfig.set(**configs)

    if scout_apm.core.install():
        config.add_tween("scout_apm.pyramid.instruments")


def instruments(handler, registry):
    def scout_tween(request):
        try:
            tr = TrackedRequest.instance()
            span = tr.start_span(operation="Controller/Pyramid")

            if ignore_path(request.path):
                tr.tag("ignore_transaction", True)

            # Capture what we can from the request, but never fail
            try:
                Context.add("path", request.path)
                Context.add("user_ip", request.remote_addr)
            except Exception:
                pass

            try:
                response = handler(request)
            except Exception:
                tr.tag("error", "true")
                raise

            # This happens further down the call chain. So time it starting
            # above, but only name it if it gets to here.
            if request.matched_route is not None:
                tr.mark_real_request()
                span.operation = "Controller/" + request.matched_route.name

        finally:
            tr.stop_span()

        return response

    return scout_tween



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from scout_apm.api.context import Context
from scout_apm.core.ignore import ignore_path
from scout_apm.core.remote_ip import RemoteIp
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


class MiddlewareTimingMiddleware(object):
    """
    Insert as early into the Middleware stack as possible (outermost layers),
    so that other middlewares called after can be timed.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        operation = "Middleware"

        TrackedRequest.instance().start_span(operation=operation)
        try:
            response = self.get_response(request)
        finally:
            TrackedRequest.instance().stop_span()
        return response


class ViewTimingMiddleware(object):
    """
    Insert as deep into the middleware stack as possible, ideally wrapping no
    other middleware. Designed to time the View itself
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """
        Wrap a single incoming request with start and stop calls.
        This will start timing, but relies on the process_view callback to
        capture more details about what view was really called, and other
        similar info.

        If process_view isn't called, then the request will not
        be recorded.  This can happen if a middleware further along the stack
        doesn't call onward, and instead returns a response directly.
        """

        tr = TrackedRequest.instance()
        tr.mark_real_request()

        # This operation name won't be recorded unless changed later in
        # process_view
        operation = "Unknown"
        tr.start_span(operation=operation)
        try:
            response = self.get_response(request)
        finally:
            tr.stop_span()
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        """
        Capture details about the view_func that is about to execute
        """
        try:
            if ignore_path(request.path):
                TrackedRequest.instance().tag("ignore_transaction", True)

            view_name = request.resolver_match._func_path
            span = TrackedRequest.instance().current_span()
            if span is not None:
                span.operation = "Controller/" + view_name
                Context.add("path", request.path)
                Context.add("user_ip", RemoteIp.lookup_from_headers(request.META))
                if getattr(request, "user", None) is not None:
                    Context.add("username", request.user.get_username())
        except Exception:
            pass

    def process_exception(self, request, exception):
        """
        Mark this request as having errored out

        Does not modify or catch or otherwise change the exception thrown
        """
        TrackedRequest.instance().tag("error", "true")


class OldStyleMiddlewareTimingMiddleware(object):
    """
    Insert as early into the Middleware stack as possible (outermost layers),
    so that other middlewares called after can be timed.
    """

    def process_request(self, request):
        operation = "Middleware"
        span = TrackedRequest.instance().start_span(operation=operation)
        request.scout_middleware_span = span

    def process_response(self, request, response):
        tr = TrackedRequest.instance()
        if (
            hasattr(request, "scout_middleware_span")
            and tr.current_span() == request.scout_middleware_span
        ):
            tr.stop_span()
        return response


class OldStyleViewMiddleware(object):
    def process_request(self, request):
        pass

    def process_view(self, request, view_func, view_func_args, view_func_kwargs):
        tr = TrackedRequest.instance()

        if ignore_path(request.path):
            tr.tag("ignore_transaction", True)

        view_name = request.resolver_match._func_path
        operation = "Controller/" + view_name

        span = tr.start_span(operation=operation)
        tr.mark_real_request()

        # Save the span into the request, so we can check
        # if we're matched up when stopping
        request.scout_view_span = span

        return None

    # Process Response could be called without ever having called process_view.
    # Be careful to not stop a span that never got started
    def process_response(self, request, response):
        tr = TrackedRequest.instance()
        if (
            hasattr(request, "scout_view_span")
            and tr.current_span() == request.scout_view_span
        ):
            tr.stop_span()
        return response

    def process_exception(self, request, exception):
        tr = TrackedRequest.instance()

        if (
            hasattr(request, "scout_view_span")
            and tr.current_span() == request.scout_view_span
        ):
            tr.tag("error", "true")
            tr.stop_span()

        return None



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from scout_apm.core.monkey import monkeypatch_method, unpatch_method
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


def import_Redis_and_Pipeline():
    import redis  # noqa: F401

    if redis.VERSION[0] >= 3:
        from redis import Redis  # noqa: F401
        from redis.client import Pipeline  # noqa: F401
    else:  # pragma: no cover
        from redis import StrictRedis as Redis  # noqa: F401
        from redis.client import BasePipeline as Pipeline  # noqa: F401

    return Redis, Pipeline


class Instrument(object):
    def __init__(self):
        self.installed = False

    def installable(self):
        try:
            Redis, Pipeline = import_Redis_and_Pipeline()
        except ImportError:
            logger.info("Unable to import for Redis instruments")
            return False
        if self.installed:
            logger.warn("Redis Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Redis instruments are not installable. Skipping.")
            return False

        self.installed = True

        self.patch_redis()
        self.patch_pipeline()

        logger.info("Instrumented Redis")
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Redis instruments are not installed. Skipping.")
            return False

        self.installed = False

        self.unpatch_redis()
        self.unpatch_pipeline()

    def patch_redis(self):
        try:
            Redis, _Pipeline = import_Redis_and_Pipeline()

            @monkeypatch_method(Redis)
            def execute_command(original, self, *args, **kwargs):
                try:
                    op = args[0]
                except (IndexError, TypeError):
                    op = "Unknown"

                tr = TrackedRequest.instance()
                tr.start_span(operation="Redis/{}".format(op))

                try:
                    return original(*args, **kwargs)
                finally:
                    tr.stop_span()

        except Exception as e:
            logger.warn(
                "Unable to instrument for Redis StrictRedis.execute_command: %r", e
            )

    def unpatch_redis(self):
        Redis, _Pipeline = import_Redis_and_Pipeline()

        unpatch_method(Redis, "execute_command")

    def patch_pipeline(self):
        try:
            _Redis, Pipeline = import_Redis_and_Pipeline()

            @monkeypatch_method(Pipeline)
            def execute(original, self, *args, **kwargs):
                tr = TrackedRequest.instance()
                tr.start_span(operation="Redis/MULTI")

                try:
                    return original(*args, **kwargs)
                finally:
                    tr.stop_span()

        except Exception as e:
            logger.warn("Unable to instrument for Redis BasePipeline.execute: %r", e)

    def unpatch_pipeline(self):
        _Redis, Pipeline = import_Redis_and_Pipeline()

        unpatch_method(Pipeline, "execute")



from __future__ import absolute_import, division, print_function, unicode_literals

import threading


class ThreadLocalSingleton(object):
    _instance_lock = threading.Lock()

    @classmethod
    def instance(cls, *args, **kwargs):
        with cls._instance_lock:
            # The thread local dict doesn't exist at all.
            # Need to create both it and the instance.
            if not hasattr(cls, "_thread_lookup"):
                cls.__new_instance(args, kwargs)
                return cls._thread_lookup.instance

            # Somehow we have the thread local dict, but failed
            # to make the instance. Just reinitialize both.
            if not hasattr(cls._thread_lookup, "instance"):
                cls.__new_instance(args, kwargs)
                return cls._thread_lookup.instance

            # The instance is None, so we'll need to recreate it
            # (and the thread local dict too).
            if cls._thread_lookup.instance is None:
                cls.__new_instance(args, kwargs)
                return cls._thread_lookup.instance

            # We made it through the checks, return the instance
            # that we know exists.
            return cls._thread_lookup.instance

    # Releasing is under the same lock as creating,
    # so it shouldn't step on each other.
    def release(self):
        with self.__class__._instance_lock:
            if getattr(self.__class__._thread_lookup, "instance", None) is self:
                self.__class__._thread_lookup.instance = None

    @classmethod
    def __new_instance(cls, *args, **kwargs):
        cls._thread_lookup = threading.local()
        cls._thread_lookup.instance = cls(args, kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from celery.signals import task_postrun, task_prerun

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest


# TODO: Capture queue.
# https://stackoverflow.com/questions/22385297/how-to-get-the-queue-in-which-a-task-was-run-celery?utm_medium=organic&utm_source=google_rich_qa&utm_campaign=google_rich_qa
def prerun_callback(sender=None, headers=None, body=None, **kwargs):
    name = kwargs["task"].name

    tr = TrackedRequest.instance()
    tr.mark_real_request()
    span = tr.start_span(operation=("Job/" + name))
    span.tag("queue", "default")


def postrun_callback(sender=None, headers=None, body=None, **kwargs):
    tr = TrackedRequest.instance()
    tr.stop_span()


def install():
    installed = scout_apm.core.install()
    if not installed:
        return

    task_prerun.connect(prerun_callback)
    task_postrun.connect(postrun_callback)


def uninstall():
    task_prerun.disconnect(prerun_callback)
    task_postrun.disconnect(postrun_callback)



from __future__ import absolute_import, division, print_function, unicode_literals

import importlib
import logging

from scout_apm.core.context import AgentContext

logger = logging.getLogger(__name__)


class InstrumentManager(object):
    INSTRUMENT_NAMESPACE = "scout_apm.instruments"
    DEFAULT_INSTRUMENTS = ["jinja2", "pymongo", "redis", "urllib3", "elasticsearch"]

    def install(self, module_name, klass="Instrument", *args, **kwargs):
        try:
            installable = importlib.import_module(module_name)
            installable = getattr(installable, klass)
            installable = installable(*args, **kwargs)

            result = getattr(installable, "install")()
            return result
        except Exception as e:
            logger.info("Exception while installing instrument %s: %r", module_name, e)
            return False

    def install_all(self):
        for module_name in self.__class__.DEFAULT_INSTRUMENTS:
            if self.is_disabled(module_name):
                logger.info("%s instruments are disabled. Skipping.", module_name)
                continue
            self.install(
                "{}.{}".format(self.__class__.INSTRUMENT_NAMESPACE, module_name)
            )

    def is_disabled(self, module_name):
        disabled = AgentContext.instance.config.value("disabled_instruments")
        if module_name in disabled:
            return True



from __future__ import absolute_import, division, print_function, unicode_literals

from scout_apm.core.config import ScoutConfig
from scout_apm.core.socket import CoreAgentSocket


class AgentContext(object):
    instance = None

    def __init__(self, *args, **kwargs):
        self.config = kwargs.get("config", ScoutConfig())
        self.config.log()

    @classmethod
    def build(cls, *args, **kwargs):
        cls.instance = AgentContext(*args, **kwargs)
        return cls.instance

    @classmethod
    def socket(cls):
        return CoreAgentSocket.instance(scout_config=ScoutConfig())



from __future__ import absolute_import, division, print_function, unicode_literals

from flask import current_app
from flask.globals import _request_ctx_stack

import scout_apm.core
from scout_apm.core.config import ScoutConfig
from scout_apm.core.ignore import ignore_path
from scout_apm.core.monkey import CallableProxy
from scout_apm.core.tracked_request import TrackedRequest


class ScoutApm(object):
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

        app.before_first_request(self.before_first_request)
        app.before_request(self.process_request)
        app.after_request(self.process_response)

        # Monkey-patch the Flask.dispatch_request method
        app.dispatch_request = self.dispatch_request

    #############
    #  Startup  #
    #############

    def before_first_request(self):
        self.extract_flask_settings()
        scout_apm.core.install()

    def extract_flask_settings(self):
        """
        Copies SCOUT_* settings in the app into Scout's config lookup
        """
        configs = {}
        configs["application_root"] = self.app.instance_path
        for name in current_app.config:
            if name.startswith("SCOUT_"):
                value = current_app.config[name]
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        ScoutConfig.set(**configs)

    #############################
    #  Request Lifecycle hooks  #
    #############################

    def dispatch_request(self):
        """Modified version of Flask.dispatch_request to call process_view."""

        req = _request_ctx_stack.top.request
        app = current_app

        # Return flask's default options response. See issue #40
        if req.method == "OPTIONS":
            return app.make_default_options_response()

        if req.routing_exception is not None:
            app.raise_routing_exception(req)

        # The routing rule has some handy attributes to extract how Flask found
        # this endpoint
        rule = req.url_rule

        # Wrap the real view_func
        view_func = self.wrap_view_func(
            app, rule, req, app.view_functions[rule.endpoint], req.view_args
        )

        return view_func(**req.view_args)

    def process_request(self):
        pass

    def wrap_view_func(self, app, rule, req, view_func, view_kwargs):
        """ This method is called just before the flask view is called.
        This is done by the dispatch_request method.
        """
        operation = view_func.__module__ + "." + view_func.__name__
        return self.trace_view_function(
            view_func, ("Controller", {"path": req.path, "name": operation})
        )

    def trace_view_function(self, func, info):
        try:

            def tracing_function(original, *args, **kwargs):
                entry_type, detail = info

                operation = entry_type
                if detail["name"] is not None:
                    operation = operation + "/" + detail["name"]

                tr = TrackedRequest.instance()
                tr.mark_real_request()
                span = tr.start_span(operation=operation)

                for key in detail:
                    span.tag(key, detail[key])

                if ignore_path(detail.get("path", "")):
                    tr.tag("ignore_transaction", True)

                # And the custom View stuff
                #  request = args[0]

                # Extract headers
                #  regex = re.compile('^HTTP_')
                #  headers = dict((regex.sub('', header), value) for (header, value)
                #  in request.META.items() if header.startswith('HTTP_'))

                #  span.tag('remote_addr', request.META['REMOTE_ADDR'])

                try:
                    return original(*args, **kwargs)
                except Exception as e:
                    TrackedRequest.instance().tag("error", "true")
                    raise e
                finally:
                    TrackedRequest.instance().stop_span()

            return CallableProxy(func, tracing_function)
        except Exception:
            # If we can't wrap for any reason, just return the original
            return func

    def process_response(self, response):
        return response



from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import logging


def download(**kwargs):
    from scout_apm.core import CoreAgentManager

    core_agent_manager = CoreAgentManager()
    core_agent_manager.download()


def launch(**kwargs):
    from scout_apm.core import CoreAgentManager

    core_agent_manager = CoreAgentManager()
    core_agent_manager.launch()


def main(**kwargs):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="count"
    )

    subparsers = parser.add_subparsers(dest="subparser")
    subparsers.add_parser("download")
    subparsers.add_parser("launch")

    args = parser.parse_args()

    if args.verbose is not None:
        if args.verbose >= 2:
            logging.basicConfig(level=getattr(logging, "DEBUG", None))
        elif args.verbose == 1:
            logging.basicConfig(level=getattr(logging, "INFO", None))

    kwargs = vars(args)
    globals()[kwargs.pop("subparser")](**kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from flask_sqlalchemy import SQLAlchemy

import scout_apm.sqlalchemy
from scout_apm.core.monkey import monkeypatch_method


def instrument_sqlalchemy(db):
    @monkeypatch_method(SQLAlchemy)
    def get_engine(original, self, *args, **kwargs):
        engine = original(*args, **kwargs)
        scout_apm.sqlalchemy.instrument_sqlalchemy(engine)
        return engine



from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib
import json
import logging
import os
import subprocess
import tarfile
import time

import requests

from scout_apm.core.context import AgentContext

logger = logging.getLogger(__name__)


class CoreAgentManager(object):
    def __init__(self):
        self.core_agent_bin_path = None
        self.core_agent_bin_version = None
        self.core_agent_dir = "{}/{}".format(
            AgentContext.instance.config.value("core_agent_dir"),
            AgentContext.instance.config.value("core_agent_full_name"),
        )
        self.downloader = CoreAgentDownloader(
            self.core_agent_dir,
            AgentContext.instance.config.value("core_agent_full_name"),
        )

    def launch(self):
        if not AgentContext.instance.config.value("core_agent_launch"):
            logger.debug(
                "Not attempting to launch Core Agent "
                "due to 'core_agent_launch' setting."
            )
            return False

        if not self.verify():
            if not AgentContext.instance.config.value("core_agent_download"):
                logger.debug(
                    "Not attempting to download Core Agent due "
                    "to 'core_agent_download' setting."
                )
                return False

            self.download()

            if not self.verify():
                logger.debug("Failed to verify Core Agent. Not launching Core Agent.")
                return False

        return self.run()

    def download(self):
        self.downloader.download()

    def run(self):
        try:
            subprocess.check_call(
                self.agent_binary()
                + self.daemonize_flag()
                + self.log_level()
                + self.log_file()
                + self.config_file()
                + self.socket_path()
            )
        except Exception as e:
            # TODO detect failure of launch properly
            logger.error("Error running Core Agent: %r", e)
            return False
        return True

    def agent_binary(self):
        return [self.core_agent_bin_path, "start"]

    def daemonize_flag(self):
        return ["--daemonize", "true"]

    def socket_path(self):
        socket_path = AgentContext.instance.config.value("socket_path")
        return ["--socket", socket_path]

    def log_level(self):
        level = AgentContext.instance.config.value("log_level")
        return ["--log-level", level]

    def log_file(self):
        path = AgentContext.instance.config.value("log_file")
        if path is not None:
            return ["--log-file", path]
        else:
            return []

    def config_file(self):
        path = AgentContext.instance.config.value("config_file")
        if path is not None:
            return ["--config-file", path]
        else:
            return []

    def verify(self):
        manifest = CoreAgentManifest(self.core_agent_dir + "/manifest.json")
        if not manifest.is_valid():
            logger.debug(
                "Core Agent verification failed: CoreAgentManifest is not valid."
            )
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False

        bin_path = self.core_agent_dir + "/" + manifest.bin_name
        if SHA256.digest(bin_path) == manifest.sha256:
            self.core_agent_bin_path = bin_path
            self.core_agent_bin_version = manifest.bin_version
            return True
        else:
            logger.debug("Core Agent verification failed: SHA mismatch.")
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False


class CoreAgentDownloader(object):
    def __init__(self, download_destination, core_agent_full_name):
        self.stale_download_secs = 120
        self.destination = download_destination
        self.core_agent_full_name = core_agent_full_name
        self.package_location = self.destination + "/{}.tgz".format(
            self.core_agent_full_name
        )
        self.download_lock_path = self.destination + "/download.lock"
        self.download_lock_fd = None

    def download(self):
        self.create_core_agent_dir()
        self.obtain_download_lock()
        if self.download_lock_fd is not None:
            try:
                self.download_package()
                self.untar()
            except OSError as e:
                logger.error("Exception raised while downloading Core Agent: %r", e)
            finally:
                self.release_download_lock()

    def create_core_agent_dir(self):
        try:
            os.makedirs(
                self.destination, AgentContext.instance.config.core_agent_permissions()
            )
        except OSError:
            pass

    def obtain_download_lock(self):
        self.clean_stale_download_lock()
        try:
            self.download_lock_fd = os.open(
                self.download_lock_path,
                os.O_RDWR | os.O_CREAT | os.O_EXCL | os.O_NONBLOCK,
            )
        except OSError as e:
            logger.debug(
                "Could not obtain download lock on %s: %r", self.download_lock_path, e
            )
            self.download_lock_fd = None

    def clean_stale_download_lock(self):
        try:
            delta = time.time() - os.stat(self.download_lock_path).st_ctime
            if delta > self.stale_download_secs:
                logger.debug("Clearing stale download lock file.")
                os.unlink(self.download_lock_path)
        except OSError:
            pass

    def release_download_lock(self):
        if self.download_lock_fd is not None:
            os.unlink(self.download_lock_path)
            os.close(self.download_lock_fd)

    def download_package(self):
        logger.debug("Downloading: %s to %s", self.full_url(), self.package_location)
        req = requests.get(self.full_url(), stream=True)
        with open(self.package_location, "wb") as f:
            for chunk in req.iter_content(1024 * 1000):
                f.write(chunk)

    def untar(self):
        t = tarfile.open(self.package_location, "r")
        t.extractall(self.destination)

    def full_url(self):
        return "{root_url}/{core_agent_full_name}.tgz".format(
            root_url=self.root_url(), core_agent_full_name=self.core_agent_full_name
        )

    def root_url(self):
        return AgentContext.instance.config.value("download_url")


class CoreAgentManifest(object):
    def __init__(self, path):
        self.manifest_path = path
        self.bin_name = None
        self.bin_version = None
        self.sha256 = None
        self.valid = False
        try:
            self.parse()
        except (ValueError, TypeError, OSError, IOError) as e:
            logger.debug("Error parsing Core Agent Manifest: %r", e)

    def parse(self):
        logger.debug("Parsing Core Agent manifest path: %s", self.manifest_path)
        with open(self.manifest_path) as manifest_file:
            self.raw = manifest_file.read()
            self.json = json.loads(self.raw)
            self.version = self.json["version"]
            self.bin_version = self.json["core_agent_version"]
            self.bin_name = self.json["core_agent_binary"]
            self.sha256 = self.json["core_agent_binary_sha256"]
            self.valid = True
            logger.debug("Core Agent manifest json: %s", self.json)

    def is_valid(self):
        return self.valid


class SHA256(object):
    @staticmethod
    def digest(filename, block_size=65536):
        try:
            sha256 = hashlib.sha256()
            with open(filename, "rb") as f:
                for block in iter(lambda: f.read(block_size), b""):
                    sha256.update(block)
            return sha256.hexdigest()
        except OSError as e:
            logger.debug("Error on digest: %r", e)
            return None



# Module for helper functions to capture tracebacks

from __future__ import absolute_import, division, print_function, unicode_literals

import traceback


def capture():
    stack = traceback.extract_stack()
    formatted_stack = []
    for frame in stack:
        # Python 2.7 and 3.4 returned tuples
        if type(frame) is tuple:
            filename = frame[0]
            line = frame[1]
            function = frame[3]
        # 3.5+ returned objects
        else:
            filename = frame.filename
            line = frame.lineno
            function = frame.name

        formatted_stack.append({"file": filename, "line": line, "function": function})

    # Python puts the closest stack frames at the end of the traceback. But we
    # want them up front
    formatted_stack.reverse()

    return formatted_stack



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

# Used in the exec() call below.
from scout_apm.core.monkey import monkeypatch_method, unpatch_method
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


class Instrument(object):
    CLIENT_METHODS = [
        "bulk",
        "count",
        "create",
        "delete",
        "delete_by_query",
        "exists",
        "exists_source",
        "explain",
        "field_caps",
        "get",
        "get_source",
        "index",
        "mget",
        "msearch",
        "msearch_template",
        "mtermvectors",
        "reindex",
        "reindex_rethrottle",
        "search",
        "search_shards",
        "search_template",
        "termvectors",
        "update",
        "update_by_query",
    ]

    def __init__(self):
        self.installed = False

    def installable(self):
        try:
            from elasticsearch import Elasticsearch  # noqa: F401
            from elasticsearch import Transport  # noqa: F401
        except ImportError:
            logger.info("Unable to import for Elasticsearch instruments")
            return False
        if self.installed:
            logger.warn("Elasticsearch Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Elasticsearch instruments are not installable. Skipping.")
            return False

        self.installed = True

        self.instrument_client()
        self.instrument_transport()

    def uninstall(self):
        if not self.installed:
            logger.info("Elasticsearch instruments are not installed. Skipping.")
            return False

        self.installed = False

        self.uninstrument_client()
        self.uninstrument_transport()

    def instrument_client(self):
        try:
            from elasticsearch import Elasticsearch  # noqa: F401
        except ImportError:
            logger.info(
                "Unable to import for Elasticsearch Client instruments. "
                "Instrument install failed."
            )
            return False

        for method_str in self.__class__.CLIENT_METHODS:
            try:
                code_str = """\
@monkeypatch_method(Elasticsearch)
def {method_str}(original, self, *args, **kwargs):
    tr = TrackedRequest.instance()
    index = kwargs.get('index', 'Unknown')
    if isinstance(index, (list, tuple)):
        index = ','.join(index)
    index = index.title()
    name = '/'.join(['Elasticsearch', index, '{camel_name}'])
    tr.start_span(operation=name, ignore_children=True)


    try:
        return original(*args, **kwargs)
    finally:
        tr.stop_span()
""".format(
                    method_str=method_str,
                    camel_name="".join(c.title() for c in method_str.split("_")),
                )

                exec(code_str)
                logger.info("Instrumented Elasticsearch Elasticsearch.%s", method_str)

            except Exception as e:
                logger.warn(
                    "Unable to instrument for Elasticsearch Elasticsearch.%s: %r",
                    method_str,
                    e,
                )
                return False
        return True

    def uninstrument_client(self):
        from elasticsearch import Elasticsearch

        for method_str in self.__class__.CLIENT_METHODS:
            unpatch_method(Elasticsearch, method_str)

    def instrument_transport(self):
        try:
            from elasticsearch import Transport

            def _sanitize_name(name):
                try:
                    op = name.split("/")[-1]
                    op = op[1:]  # chop leading '_' from op
                    allowed_names = [
                        "bench",
                        "bulk",
                        "count",
                        "exists",
                        "explain",
                        "field_stats",
                        "health",
                        "mget",
                        "mlt",
                        "mpercolate",
                        "msearch",
                        "mtermvectors",
                        "percolate",
                        "query",
                        "scroll",
                        "search_shards",
                        "source",
                        "suggest",
                        "template",
                        "termvectors",
                        "update",
                        "search",
                    ]

                    if op in allowed_names:
                        return op.title()
                    return "Unknown"
                except Exception:
                    return "Unknown"

            @monkeypatch_method(Transport)
            def perform_request(original, self, *args, **kwargs):
                try:
                    op = _sanitize_name(args[1])
                except IndexError:
                    op = "Unknown"

                tr = TrackedRequest.instance()
                tr.start_span(
                    operation="Elasticsearch/{}".format(op), ignore_children=True
                )

                try:
                    return original(*args, **kwargs)
                finally:
                    tr.stop_span()

            logger.info("Instrumented Elasticsearch Transport")

        except Exception as e:
            logger.warn(
                "Unable to instrument for Elasticsearch Transport.perform_request: %r",
                e,
            )
            return False
        return True

    def uninstrument_transport(self):
        from elasticsearch import Transport

        unpatch_method(Transport, "perform_request")



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import sys

import scout_apm.core
from scout_apm.api.context import Context as ScoutContext
from scout_apm.core.config import ScoutConfig
from scout_apm.core.tracked_request import TrackedRequest

logger = logging.getLogger(__name__)


# Python 2 (and very early 3.x) didn't have ContextDecorator, so define it for ourselves
if sys.version_info < (3, 2):
    import functools

    class ContextDecorator(object):
        def __call__(self, f):
            @functools.wraps(f)
            def decorated(*args, **kwds):
                with self:
                    return f(*args, **kwds)

            return decorated


else:
    from contextlib import ContextDecorator


text_type = str if sys.version_info[0] >= 3 else unicode  # noqa: F821


def text(value, encoding="utf-8", errors="strict"):
    """Convert a value to str on Python 3 and unicode on Python 2."""
    if isinstance(value, text_type):
        return value
    elif isinstance(value, bytes):
        return text_type(value, encoding, errors)
    else:
        return text_type(value)


class Context(ScoutContext):
    pass


class Config(ScoutConfig):
    pass


def install(*args, **kwargs):
    scout_apm.core.install(*args, **kwargs)


def ignore_transaction():
    TrackedRequest.instance().tag("ignore_transaction", True)


class instrument(ContextDecorator):
    def __init__(self, operation, kind="Custom", tags={}):
        self.operation = text(kind) + "/" + text(operation)
        self.tags = tags

    def __enter__(self):
        tr = TrackedRequest.instance()
        self.span = tr.start_span(operation=self.operation)
        for key, value in self.tags.items():
            self.tag(key, value)
        return self

    def __exit__(self, *exc):
        tr = TrackedRequest.instance()
        tr.stop_span()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class Transaction(ContextDecorator):
    """
    This Class is not meant to be used directly.
    Use one of the subclasses
    (WebTransaction or BackgroundTransaction)
    """

    def __init__(self, name, tags={}):
        self.name = text(name)
        self.tags = tags

    @classmethod
    def start(cls, kind, name, tags={}):
        operation = text(kind) + "/" + text(name)

        tr = TrackedRequest.instance()
        tr.mark_real_request()
        span = tr.start_span(operation=operation)
        for key, value in tags.items():
            tr.tag(key, value)
        return span

    @classmethod
    def stop(cls):
        tr = TrackedRequest.instance()
        tr.stop_span()
        return True

    # __enter__ must be defined by child classes.

    # *exc is any exception raised. Ignore that
    def __exit__(self, *exc):
        WebTransaction.stop()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class WebTransaction(Transaction):
    @classmethod
    def start(cls, name, tags={}):
        Transaction.start("Controller", text(name), tags)

    def __enter__(self):
        Transaction.start("Controller", self.name, self.tags)


class BackgroundTransaction(Transaction):
    @classmethod
    def start(cls, name, tags={}):
        Transaction.start("Job", text(name), tags)

    def __enter__(self):
        Transaction.start("Job", self.name, self.tags)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from datetime import datetime
from uuid import uuid4

import scout_apm.core.backtrace
from scout_apm.core.n_plus_one_call_set import NPlusOneCallSet
from scout_apm.core.request_manager import RequestManager
from scout_apm.core.samplers import Memory, Samplers
from scout_apm.core.thread_local import ThreadLocalSingleton

try:
    from scout_apm.core import objtrace
except ImportError:
    objtrace = None

logger = logging.getLogger(__name__)


class TrackedRequest(ThreadLocalSingleton):
    """
    This is a container which keeps track of all module instances for a single
    request. For convenience they are made available as attributes based on
    their keyname
    """

    def __init__(self, *args, **kwargs):
        self.req_id = "req-" + str(uuid4())
        self.start_time = kwargs.get("start_time", datetime.utcnow())
        self.end_time = kwargs.get("end_time", None)
        self.active_spans = kwargs.get("active_spans", [])
        self.complete_spans = kwargs.get("complete_spans", [])
        self.tags = kwargs.get("tags", {})
        self.real_request = kwargs.get("real_request", False)
        self.memory_start = kwargs.get("memory_start", Memory.rss_in_mb())
        self.callset = NPlusOneCallSet()
        logger.debug("Starting request: %s", self.req_id)

    def mark_real_request(self):
        self.real_request = True

    def is_real_request(self):
        return self.real_request

    def tag(self, key, value):
        if key in self.tags:
            logger.debug(
                "Overwriting previously set tag for request %s: %s", self.req_id, key
            )
        self.tags[key] = value

    def start_span(self, *args, **kwargs):
        maybe_parent = self.current_span()

        if maybe_parent is not None:
            parent_id = maybe_parent.span_id
            if maybe_parent.ignore_children:
                kwargs["ignore"] = True
                kwargs["ignore_children"] = True
        else:
            parent_id = None

        kwargs["parent"] = parent_id
        kwargs["request_id"] = self.req_id

        new_span = Span(**kwargs)
        self.active_spans.append(new_span)
        return new_span

    def stop_span(self):
        try:
            stopping_span = self.active_spans.pop()
        except IndexError as e:
            logger.debug("Exception when stopping span: %r", e)
        else:
            stopping_span.stop()
            if not stopping_span.ignore:
                stopping_span.annotate()
                self.complete_spans.append(stopping_span)

        if len(self.active_spans) == 0:
            self.finish()

    def current_span(self):
        if len(self.active_spans) > 0:
            return self.active_spans[-1]
        else:
            return None

    # Request is done, release any info we have about it.
    def finish(self):
        logger.debug("Stopping request: %s", self.req_id)
        if self.end_time is None:
            self.end_time = datetime.utcnow()
        if self.is_real_request():
            self.tag("mem_delta", Memory.get_delta(self.memory_start))
            if not self.is_ignored():
                RequestManager.instance().add_request(self)
            Samplers.ensure_running()

        # This can fail if the Tracked Request was created directly,
        # not through instance()
        try:
            self.release()
        except Exception:
            pass

    # A request is ignored if the tag "ignore_transaction" is set to True
    def is_ignored(self):
        return self.tags.get("ignore_transaction", False)


class Span(object):
    def __init__(self, *args, **kwargs):
        self.span_id = kwargs.get("span_id", "span-" + str(uuid4()))
        self.start_time = kwargs.get("start_time", datetime.utcnow())
        self.end_time = kwargs.get("end_time", None)
        self.request_id = kwargs.get("request_id", None)
        self.operation = kwargs.get("operation", None)
        self.ignore = kwargs.get("ignore", False)
        self.ignore_children = kwargs.get("ignore_children", False)
        self.parent = kwargs.get("parent", None)
        self.tags = kwargs.get("tags", {})
        if objtrace is not None:
            self.start_objtrace_counts = kwargs.get(
                "start_objtrace_counts", objtrace.get_counts()
            )
        else:
            self.start_objtrace_counts = kwargs.get(
                "start_objtrace_counts", (0, 0, 0, 0)
            )
        self.end_objtrace_counts = kwargs.get("end_objtrace_counts", (0, 0, 0, 0))

    def stop(self):
        self.end_time = datetime.utcnow()
        if objtrace is not None:
            self.end_objtrace_counts = objtrace.get_counts()
        else:
            self.end_objtrace_counts = (0, 0, 0, 0)

    def tag(self, key, value):
        if key in self.tags:
            logger.debug(
                "Overwriting previously set tag for span %s: %s", self.span_id, key
            )
        self.tags[key] = value

    # In seconds
    def duration(self):
        if self.end_time is not None:
            return (self.end_time - self.start_time).total_seconds()
        else:
            # Current, running duration
            return (datetime.utcnow() - self.start_time).total_seconds()

    # Add any interesting annotations to the span. Assumes that we are in the
    # process of stopping this span.
    def annotate(self):
        self.tag("allocations", self.calculate_allocations())
        # Don't capture backtraces for Controller or Middleware
        if self.operation is not None:
            if self.operation.startswith("Controller") or self.operation.startswith(
                "Middleware"
            ):
                return
        slow_threshold = 0.500
        if self.duration() > slow_threshold:
            self.capture_backtrace()

    def calculate_allocations(self):
        if objtrace is None:
            return 0

        start_allocs = (
            self.start_objtrace_counts[0]
            + self.start_objtrace_counts[1]
            + self.start_objtrace_counts[2]
        )
        end_allocs = (
            self.end_objtrace_counts[0]
            + self.end_objtrace_counts[1]
            + self.end_objtrace_counts[2]
        )
        try:
            # If even one of the counters rolled over, we're pretty much
            # guaranteed to have end_allocs be less than start_allocs.
            # This should rarely happen. Max Unsigned Long Long is a big number
            if end_allocs - start_allocs < 0:
                logger.debug(
                    "End allocation count smaller than start allocation "
                    "count for span %s: start = %d, end = %d",
                    self.span_id,
                    start_allocs,
                    end_allocs,
                )
                return 0
            return end_allocs - start_allocs
        except TypeError as e:
            logger.debug("Exception in calculate_allocations: %r", e)
            return 0

    def capture_backtrace(self):
        stack = scout_apm.core.backtrace.capture()
        self.tag("stack", stack)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from django.conf import settings

from scout_apm.core.config import ScoutConfig

logger = logging.getLogger(__name__)


class ConfigAdapter(object):
    @classmethod
    def install(cls):
        configs = {}
        if getattr(settings, "BASE_DIR", None) is not None:
            configs["application_root"] = settings.BASE_DIR
        for name in dir(settings):
            if name.startswith("SCOUT_"):
                value = getattr(settings, name)
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        ScoutConfig.set(**configs)



# flake8: noqa

# Originally taken from https://pypi.python.org/pypi/ProxyTypes
# inlined due to python3 issues with setup.py, we should gut this out and simplify it


class AbstractProxy(object):
    """Delegates all operations (except ``.__subject__``) to another object"""

    __slots__ = ()

    def __call__(self, *args, **kw):
        return self.__subject__(*args, **kw)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        subject = oga(self, "__subject__")
        if attr == "__subject__":
            return subject
        return getattr(subject, attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if attr == "__subject__":
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if attr == "__subject__":
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)

    def __nonzero__(self):
        return bool(self.__subject__)

    def __getitem__(self, arg):
        return self.__subject__[arg]

    def __setitem__(self, arg, val):
        self.__subject__[arg] = val

    def __delitem__(self, arg):
        del self.__subject__[arg]

    def __getslice__(self, i, j):
        return self.__subject__[i:j]

    def __setslice__(self, i, j, val):
        self.__subject__[i:j] = val

    def __delslice__(self, i, j):
        del self.__subject__[i:j]

    def __contains__(self, ob):
        return ob in self.__subject__

    for name in "repr str hash len abs complex int long float iter oct hex".split():
        exec("def __%s__(self): return %s(self.__subject__)" % (name, name))

    for name in "cmp", "coerce", "divmod":
        exec("def __%s__(self,ob): return %s(self.__subject__,ob)" % (name, name))

    for name, op in [
        ("lt", "<"),
        ("gt", ">"),
        ("le", "<="),
        ("ge", ">="),
        ("eq", "=="),
        ("ne", "!="),
    ]:
        exec("def __%s__(self,ob): return self.__subject__ %s ob" % (name, op))

    for name, op in [("neg", "-"), ("pos", "+"), ("invert", "~")]:
        exec("def __%s__(self): return %s self.__subject__" % (name, op))

    for name, op in [
        ("or", "|"),
        ("and", "&"),
        ("xor", "^"),
        ("lshift", "<<"),
        ("rshift", ">>"),
        ("add", "+"),
        ("sub", "-"),
        ("mul", "*"),
        ("div", "/"),
        ("mod", "%"),
        ("truediv", "/"),
        ("floordiv", "//"),
    ]:
        exec(
            (
                "def __%(name)s__(self,ob):\n"
                "    return self.__subject__ %(op)s ob\n"
                "\n"
                "def __r%(name)s__(self,ob):\n"
                "    return ob %(op)s self.__subject__\n"
                "\n"
                "def __i%(name)s__(self,ob):\n"
                "    self.__subject__ %(op)s=ob\n"
                "    return self\n"
            )
            % locals()
        )

    del name, op

    # Oddball signatures

    def __rdivmod__(self, ob):
        return divmod(ob, self.__subject__)

    def __pow__(self, *args):
        return pow(self.__subject__, *args)

    def __ipow__(self, ob):
        self.__subject__ **= ob
        return self

    def __rpow__(self, ob):
        return pow(ob, self.__subject__)


class ObjectProxy(AbstractProxy):
    """Proxy for a specific object"""

    __slots__ = "__subject__"

    def __init__(self, subject):
        self.__subject__ = subject


class CallbackProxy(AbstractProxy):
    """Proxy for a dynamically-chosen object"""

    __slots__ = "__callback__"

    def __init__(self, func):
        set_callback(self, func)


set_callback = CallbackProxy.__callback__.__set__
get_callback = CallbackProxy.__callback__.__get__
CallbackProxy.__subject__ = property(lambda self, gc=get_callback: gc(self)())


class LazyProxy(CallbackProxy):
    """Proxy for a lazily-obtained object, that is cached on first use"""

    __slots__ = "__cache__"


get_cache = LazyProxy.__cache__.__get__
set_cache = LazyProxy.__cache__.__set__


def __subject__(self, get_cache=get_cache, set_cache=set_cache):
    try:
        return get_cache(self)
    except AttributeError:
        set_cache(self, get_callback(self)())
        return get_cache(self)


LazyProxy.__subject__ = property(__subject__, set_cache)
del __subject__


class AbstractWrapper(AbstractProxy):
    """Mixin to allow extra behaviors and attributes on proxy instance"""

    __slots__ = ()

    def __getattribute__(self, attr, oga=object.__getattribute__):
        if attr.startswith("__"):
            subject = oga(self, "__subject__")
            if attr == "__subject__":
                return subject
            return getattr(subject, attr)
        return oga(self, attr)

    def __getattr__(self, attr, oga=object.__getattribute__):
        return getattr(oga(self, "__subject__"), attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if (
            attr == "__subject__"
            or hasattr(type(self), attr)
            and not attr.startswith("__")
        ):
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if (
            attr == "__subject__"
            or hasattr(type(self), attr)
            and not attr.startswith("__")
        ):
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)


class ObjectWrapper(ObjectProxy, AbstractWrapper):
    __slots__ = ()


class CallbackWrapper(CallbackProxy, AbstractWrapper):
    __slots__ = ()


class LazyWrapper(LazyProxy, AbstractWrapper):
    __slots__ = ()


#######################
#  MONKEYS DOWN HERE  #
#######################


class CallableProxy(ObjectWrapper):
    __slots__ = "_eop_wrapper_"

    def __init__(self, wrapped, wrapper):
        super(CallableProxy, self).__init__(wrapped)
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(self.__subject__, *args, **kwargs)


class BoundMethodProxy(ObjectWrapper):
    __slots__ = ("_eop_wrapper_", "_eop_instance_")

    def __init__(self, wrapped, instance, wrapper):
        super(BoundMethodProxy, self).__init__(wrapped)
        self._eop_instance_ = instance
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(
            self.__subject__, self._eop_instance_, *args, **kwargs
        )


class UnboundMethodProxy(CallableProxy):
    __slots__ = "_eop_wrapper_"

    def __get__(self, instance, owner):
        return BoundMethodProxy(
            self.__subject__.__get__(instance, owner),
            instance or owner,
            self._eop_wrapper_,
        )

    def __getattribute__(self, attr, oga=object.__getattribute__):
        """
        We need to return our own version of __get__ or we may end up
        never being called if the member we are wrapping is wrapped
        again by someone else
        """
        if attr == "__get__":
            return oga(self, attr)
        return super(UnboundMethodProxy, self).__getattribute__(attr)


def monkeypatch_method(cls, method_name=None):
    def decorator(func):
        method_to_patch = method_name or func.__name__
        original = cls.__dict__[method_to_patch]
        replacement = UnboundMethodProxy(original, func)
        type.__setattr__(cls, method_to_patch, replacement)  # Avoid overrides
        return func

    return decorator


# Slightly annoying signature because there are no unbound methods in Python 3.
def unpatch_method(cls, method_name):
    type.__setattr__(cls, method_name, getattr(cls, method_name).__subject__)



from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
from datetime import datetime

logger = logging.getLogger(__name__)

# A constant that represents an "unknown" date. Fallback, and will be rejected
# by the CoreAgent. But important to avoid exceptions if a None timestamp is
# passed to a Command.
INVALID_DATE = datetime(year=2000, month=1, day=1)


class Register(object):
    def __init__(self, *args, **kwargs):
        self.app = kwargs.get("app")
        self.key = kwargs.get("key")

    def message(self):
        logger.info("Registering with app=%s key=%s" % (self.app, self.key))
        return {
            "Register": {
                "app": self.app,
                "key": self.key,
                "language": "python",
                "api_version": "1.0",
            }
        }


class StartSpan(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")
        self.span_id = kwargs.get("span_id")
        self.parent = kwargs.get("parent")
        self.operation = kwargs.get("operation")

    def message(self):
        return {
            "StartSpan": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
                "span_id": self.span_id,
                "parent_id": self.parent,
                "operation": self.operation,
            }
        }


class StopSpan(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")
        self.span_id = kwargs.get("span_id")

    def message(self):
        return {
            "StopSpan": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
                "span_id": self.span_id,
            }
        }


class StartRequest(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")

    def message(self):
        return {
            "StartRequest": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
            }
        }


class FinishRequest(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")

    def message(self):
        return {
            "FinishRequest": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
            }
        }


class TagSpan(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")
        self.span_id = kwargs.get("span_id")
        self.tag = kwargs.get("tag")
        self.value = kwargs.get("value")

    def message(self):
        return {
            "TagSpan": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
                "span_id": self.span_id,
                "tag": self.tag,
                "value": self.value,
            }
        }


class TagRequest(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.request_id = kwargs.get("request_id")
        self.tag = kwargs.get("tag")
        self.value = kwargs.get("value")

    def message(self):
        return {
            "TagRequest": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "request_id": self.request_id,
                "tag": self.tag,
                "value": self.value,
            }
        }


class ApplicationEvent(object):
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get("timestamp") or INVALID_DATE
        self.event_type = kwargs.get("event_type", "")
        self.event_value = kwargs.get("event_value", "")
        self.source = kwargs.get("source", "")

    def message(self):
        return {
            "ApplicationEvent": {
                "timestamp": self.timestamp.isoformat() + "Z",
                "event_type": self.event_type,
                "event_value": self.event_value,
                "source": self.source,
            }
        }


class CoreAgentVersion(object):
    def message(self):
        return {"CoreAgentVersion": {}}


class CoreAgentVersionResponse(object):
    def __init__(self, message):
        self.loaded = json.loads(message)
        self.version = self.loaded["CoreAgentVersion"]["version"]


class BatchCommand(object):
    def __init__(self, commands):
        self.commands = commands

    def message(self):
        messages = [command.message() for command in self.commands]
        return {"BatchCommand": {"commands": messages}}

    @classmethod
    def from_tracked_request(cls, request):
        commands = []
        # Request Start
        commands.append(
            StartRequest(timestamp=request.start_time, request_id=request.req_id)
        )
        # Request Tags
        for key in request.tags:
            commands.append(
                TagRequest(
                    timestamp=request.start_time,
                    request_id=request.req_id,
                    tag=key,
                    value=request.tags[key],
                )
            )
        # Spans
        for span in request.complete_spans:
            # Span Start
            commands.append(
                StartSpan(
                    timestamp=span.start_time,
                    request_id=span.request_id,
                    span_id=span.span_id,
                    parent=span.parent,
                    operation=span.operation,
                )
            )
            # Span Tags
            for key in span.tags:
                commands.append(
                    TagSpan(
                        timestamp=span.start_time,
                        request_id=request.req_id,
                        span_id=span.span_id,
                        tag=key,
                        value=span.tags[key],
                    )
                )
            # Span End
            if span.end_time is None:
                logger.debug(
                    "Invalid Request, span_id: %s had a None end_time", span.span_id
                )
                return None

            commands.append(
                StopSpan(
                    timestamp=span.end_time,
                    request_id=span.request_id,
                    span_id=span.span_id,
                )
            )
        # Request Finish
        if request.end_time is None:
            logger.debug(
                "Invalid Request, request_id: %s had a None end_time", request.req_id
            )
            return None

        commands.append(
            FinishRequest(timestamp=request.end_time, request_id=request.req_id)
        )

        return BatchCommand(commands)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import sys
from datetime import datetime
from os import getpid

from scout_apm.core.commands import ApplicationEvent
from scout_apm.core.context import AgentContext

logger = logging.getLogger(__name__)


class AppMetadata(object):
    @classmethod
    def report(cls):
        event = ApplicationEvent()
        event.event_value = cls.data()
        event.event_type = "scout.metadata"
        event.timestamp = datetime.utcnow()
        event.source = "Pid: " + str(getpid())
        AgentContext.socket().send(event)

    @classmethod
    def data(cls):
        config = AgentContext.instance.config
        try:
            data = {
                "language": "python",
                "version": "{}.{}.{}".format(*sys.version_info[:3]),
                "server_time": datetime.utcnow().isoformat() + "Z",
                "framework": config.value("framework"),
                "framework_version": config.value("framework_version"),
                "environment": "",
                "app_server": config.value("app_server"),
                "hostname": config.value("hostname"),
                "database_engine": "",  # Detected
                "database_adapter": "",  # Raw
                "application_name": "",  # Environment.application_name,
                "libraries": cls.get_python_packages_versions(),
                "paas": "",
                "application_root": config.value("application_root"),
                "scm_subdirectory": config.value("scm_subdirectory"),
                "git_sha": config.value("revision_sha"),
            }
        except Exception as e:
            logger.debug("Exception in AppMetadata: %r", e)
            data = {}

        return data

    @classmethod
    def get_python_packages_versions(cls):
        try:
            import pkg_resources
        except ImportError:
            return []

        return list(
            sorted(
                (distribution.project_name, distribution.version)
                for distribution in pkg_resources.working_set
            )
        )



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

# Used in the exec() call below.
from scout_apm.core.monkey import monkeypatch_method, unpatch_method  # noqa: F401
from scout_apm.core.tracked_request import TrackedRequest  # noqa: F401

logger = logging.getLogger(__name__)


class Instrument(object):
    PYMONGO_METHODS = [
        "aggregate",
        "bulk_write",
        "count",
        "create_index",
        "create_indexes",
        "delete_many",
        "delete_one",
        "distinct",
        "drop",
        "drop_index",
        "drop_indexes",
        "ensure_index",
        "find_and_modify",
        "find_one",
        "find_one_and_delete",
        "find_one_and_replace",
        "find_one_and_update",
        "group",
        "inline_map_reduce",
        "insert",
        "insert_many",
        "insert_one",
        "map_reduce",
        "reindex",
        "remove",
        "rename",
        "replace_one",
        "save",
        "update",
        "update_many",
        "update_one",
    ]

    def __init__(self):
        self.installed = False

    def installable(self):
        try:
            from pymongo.collection import Collection  # noqa: F401
        except ImportError:
            logger.info("Unable to import for PyMongo instruments")
            return False
        if self.installed:
            logger.warn("PyMongo Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("PyMongo instruments are not installable. Skipping.")
            return False

        self.installed = True

        try:
            from pymongo.collection import Collection  # noqa: F401
        # There is no way the import can fail if self.installable() succeeded.
        except ImportError:  # pragma: no cover
            logger.info(
                "Unable to import for PyMongo instruments. Instrument install failed."
            )
            return False

        for method_str in self.__class__.PYMONGO_METHODS:
            try:
                code_str = """
@monkeypatch_method(Collection)
def {method_str}(original, self, *args, **kwargs):
    tr = TrackedRequest.instance()
    name = '/'.join(['MongoDB', self.name, '{camel_name}'])
    span = tr.start_span(operation=name, ignore_children=True)
    span.tag('name', self.name)

    try:
        return original(*args, **kwargs)
    finally:
        tr.stop_span()
""".format(
                    method_str=method_str,
                    camel_name="".join(c.title() for c in method_str.split("_")),
                )

                exec(code_str)
                logger.info("Instrumented PyMongo Collection.%s", method_str)

            except Exception as e:
                logger.warn(
                    "Unable to instrument for PyMongo Collection.%s: %r", method_str, e
                )
                return False
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("PyMongo instruments are not installed. Skipping.")
            return False

        self.installed = False

        from pymongo.collection import Collection

        for method_str in self.__class__.PYMONGO_METHODS:
            unpatch_method(Collection, method_str)



from __future__ import absolute_import, division, print_function, unicode_literals

import os

# Note that this is called from Config Defaults, which is looked up after the
# configs from environment variables. If you set SCOUT_REVISION_SHA, it will be
# picked up in that step of config, not here.

# Similarly, this can be overriden by the code level configuration (from Django
# configs or similar) by using the "revision_sha" key.


class GitRevision(object):
    def detect(self):
        return self.detect_from_env_var() or self.detect_from_heroku() or ""

    def detect_from_heroku(self):
        return os.environ.get("HEROKU_SLUG_COMMIT")

    def detect_from_env_var(self):
        return os.environ.get("SCOUT_REVISION_SHA")



from __future__ import absolute_import, division, print_function, unicode_literals

from scout_apm.core.tracked_request import TrackedRequest


class Context(object):
    @staticmethod
    def add(key, value):
        """Adds context to the currently executing request.

        :key: Any String identifying the request context.
              Example: "user_ip", "plan", "alert_count"
        :value: Any json-serializable type.
              Example: "1.1.1.1", "free", 100
        :returns: nothing.
        """
        tr = TrackedRequest.instance()
        tr.tag(key, value)



from __future__ import absolute_import, division, print_function, unicode_literals


# Takes an integer or a string containing an integer and returns
# the octal value.
# Raises a ValueError if the value cannot be converted to octal.
def octal(value):
    return int("{}".format(value), 8)



from __future__ import absolute_import, division, print_function, unicode_literals

import platform
import subprocess


class PlatformDetection(object):
    """
    This helps figuring out what platform we're running on, so we can download
    the correct binary build of Core Agent.
    """

    @classmethod
    def get_triple(cls):
        return "{arch}-{platform}".format(arch=cls.arch(), platform=cls.platform())

    @classmethod
    def arch(cls):
        """
        What CPU are we on?
        """
        arch = platform.machine()
        if arch == "i686":
            return "i686"
        elif arch == "x86_64":
            return "x86_64"
        else:
            return "unknown"

    @classmethod
    def platform(cls):
        """
        What Operating System (and sub-system like glibc / musl)
        """
        system_name = platform.system()
        if system_name == "Linux":
            libc = cls.libc()
            return "unknown-linux-{libc}".format(libc=libc)
        elif system_name == "Darwin":
            return "apple-darwin"
        else:
            return "unknown"

    @classmethod
    def libc(cls):
        """
        Alpine linux uses a non glibc version of the standard library, it uses
        the stripped down musl instead. The core agent can be built against it,
        but which one is running must be detected. Shelling out to `ldd`
        appears to be the most reliable way to do this.
        """
        try:
            output = subprocess.check_output(
                ["ldd", "--version"], stderr=subprocess.STDOUT
            )
        except (OSError, subprocess.CalledProcessError):
            return "gnu"
        else:
            if b"musl" in output:
                return "musl"
            else:
                return "gnu"



from __future__ import absolute_import, division, print_function, unicode_literals


class RemoteIp(object):
    """
    A helper class to lookup what IP the request is associated with for adding
    Context to a request

    This is not suitable to use in any security sensitive position. This is
    spoofable by the requester.

    Some of this algorithm is borrowed for Ruby on Rails' RemoteIp middleware
    """

    @classmethod
    def lookup_from_headers(cls, headers):
        """
        Given a dictionary of headers (WSGI request.META for instance), look up
        the most likely user's IP
        """

        # A single address, set by this server, returned as an Array
        remote_addr = cls.ips_from(headers.get("REMOTE_ADDR"))

        # Could be a CSV list and/or repeated headers that were concatenated.
        forwarded_ips = cls.ips_from(headers.get("HTTP_X_FORWARDED_FOR"))
        client_ips = cls.ips_from(headers.get("HTTP_CLIENT_IP"))

        # We assume these things about the IP headers:
        #
        #   - X-Forwarded-For will be a list of IPs, one per proxy, or blank.
        #       in order: `client,proxy1,proxy2`
        #   - Client-Ip is propagated from the outermost proxy, or is blank
        #   - REMOTE_ADDR will be the IP that made the request to this server
        #
        # X-Forwarded-For and Client-Ip shouldn't be set at the same time, but
        # if they are, use the one in Forwarded
        ips = forwarded_ips + client_ips + remote_addr

        try:
            return ips[0]
        except IndexError:
            return None

    @classmethod
    def ips_from(cls, header):
        """Splits the header into a list"""

        if header is None:
            return []

        return header.split(",")



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from os import getpid

from scout_apm.core.config import ScoutConfig
from scout_apm.core.context import AgentContext
from scout_apm.core.core_agent_manager import CoreAgentManager
from scout_apm.core.instrument_manager import InstrumentManager
from scout_apm.core.metadata import AppMetadata

try:
    from scout_apm.core import objtrace
except ImportError:
    objtrace = None

logger = logging.getLogger(__name__)


def install(*args, **kwargs):
    if "config" in kwargs:
        ScoutConfig().set(**kwargs["config"])
    context = AgentContext.build(config=ScoutConfig())

    if not context.config.value("monitor"):
        logger.info(
            "APM Not Launching on PID: %s - Configuration 'monitor' is not true",
            getpid(),
        )
        return False

    InstrumentManager().install_all()

    if objtrace is not None:
        objtrace.enable()

    logger.debug("APM Launching on PID: %s", getpid())
    launched = CoreAgentManager().launch()

    AppMetadata.report()
    if launched:
        AgentContext.socket().stop()

    return True



from __future__ import absolute_import, division, print_function, unicode_literals

from bottle import request

import scout_apm.core
from scout_apm.api.context import Context
from scout_apm.core.config import ScoutConfig
from scout_apm.core.context import AgentContext
from scout_apm.core.ignore import ignore_path
from scout_apm.core.tracked_request import TrackedRequest


class ScoutPlugin(object):
    def __init__(self):
        self.name = "scout"
        self.api = 2

    def set_config_from_bottle(self, app):
        scout_config = ScoutConfig()
        bottle_configs = {}
        for k in scout_config.known_keys():
            value = app.config.get("scout.{}".format(k))
            if value is not None and value != "":
                bottle_configs[k] = value
        scout_config.set(**bottle_configs)
        return scout_config

    def setup(self, app):
        self.set_config_from_bottle(app)
        scout_apm.core.install()

    def apply(self, callback, context):
        if not AgentContext.instance.config.value("monitor"):
            return callback

        def wrapper(*args, **kwargs):
            try:
                tr = TrackedRequest.instance()
                tr.mark_real_request()
                path = "Unknown"

                if request.route.name is not None:
                    path = request.route.name
                else:
                    path = request.route.rule

                if path == "/":
                    path = "/home"

                if not path.startswith("/"):
                    path = "/{}".format(path)

                tr.start_span(operation="Controller{}".format(path))

                if ignore_path(path):
                    tr.tag("ignore_transaction", True)

                try:
                    Context.add("path", path)
                    Context.add("user_ip", request.remote_addr)
                except Exception:
                    pass

                try:
                    response = callback(*args, **kwargs)
                except Exception:
                    tr.tag("error", "true")
                    raise

            finally:
                tr.stop_span()

            return response

        return wrapper



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import sys

from scout_apm.core.git_revision import GitRevision
from scout_apm.core.platform_detection import PlatformDetection
from scout_apm.core.util import octal

logger = logging.getLogger(__name__)


class ScoutConfig(object):
    """
    Configuration object for the ScoutApm agent.

    Contains a list of configuration "layers". When a configuration key is
    looked up, each layer is asked in turn if it knows the value. The first one
    to answer affirmatively returns the value.
    """

    def __init__(self):
        self.layers = [
            ScoutConfigEnv(),
            ScoutConfigPython(),
            ScoutConfigDerived(self),
            ScoutConfigDefaults(),
            ScoutConfigNull(),
        ]

    def value(self, key):
        value = self.locate_layer_for_key(key).value(key)
        if key in CONVERSIONS:
            converted_value = CONVERSIONS[key].convert(value)
        else:
            converted_value = value

        return converted_value

    def locate_layer_for_key(self, key):
        for layer in self.layers:
            if layer.has_config(key):
                return layer
        else:  # pragma: no cover
            # Not reachable because ScoutConfigNull returns None for all keys.
            assert False, "key not found in any layer"

    def log(self):
        logger.debug("Configuration Loaded:")
        for key in self.known_keys():
            layer = self.locate_layer_for_key(key)
            logger.debug("%-9s: %s = %s", layer.name(), key, layer.value(key))

    def known_keys(self):
        return [
            "app_server",
            "application_root",
            "core_agent_dir",
            "core_agent_download",
            "core_agent_launch",
            "core_agent_permissions",
            "core_agent_version",
            "disabled_instruments",
            "download_url",
            "framework",
            "framework_version",
            "hostname",
            "ignore",
            "key",
            "log_level",
            "monitor",
            "name",
            "revision_sha",
            "scm_subdirectory",
            "socket_path",
        ]

    def core_agent_permissions(self):
        try:
            return octal(self.value("core_agent_permissions"))
        except ValueError as e:
            logger.error(
                "Invalid core_agent_permissions value: %s." " Using default: %s",
                repr(e),
                0o700,
            )
            return 0o700

    @classmethod
    def set(cls, **kwargs):
        """
        Sets a configuration value for the Scout agent. Values set here will
        not override values set in ENV.
        """
        global SCOUT_PYTHON_VALUES
        for key, value in kwargs.items():
            SCOUT_PYTHON_VALUES[key] = value

    @classmethod
    def reset_all(cls):
        """
        Remove all configuration settings set via `ScoutConfig.set(...)`.

        This is meant for use in testing.
        """
        global SCOUT_PYTHON_VALUES
        SCOUT_PYTHON_VALUES.clear()


# Module-level data, the ScoutConfig.set(key="value") adds to this
SCOUT_PYTHON_VALUES = {}


class ScoutConfigPython(object):
    """
    A configuration overlay that lets other parts of python set values.
    """

    def name(self):
        return "Python"

    def has_config(self, key):
        return key in SCOUT_PYTHON_VALUES

    def value(self, key):
        return SCOUT_PYTHON_VALUES[key]


class ScoutConfigEnv(object):
    """
    Reads configuration from environment by prefixing the key
    requested with "SCOUT_"

    Example: the `log_level` config looks for SCOUT_LOG_LEVEL
    environment variable
    """

    def name(self):
        return "ENV"

    def has_config(self, key):
        env_key = self.modify_key(key)
        return env_key in os.environ

    def value(self, key):
        env_key = self.modify_key(key)
        return os.environ[env_key]

    def modify_key(self, key):
        env_key = ("SCOUT_" + key).upper()
        return env_key


class ScoutConfigDerived(object):
    """
    A configuration overlay that calculates from other values.
    """

    def __init__(self, config):
        """
        config argument is the overall ScoutConfig var, so we can lookup the
        components of the derived info.
        """
        self.config = config

    def name(self):
        return "Derived"

    def has_config(self, key):
        return self.lookup_func(key) is not None

    def value(self, key):
        return self.lookup_func(key)()

    def lookup_func(self, key):
        """
        Returns the derive_#{key} function, or None if it isn't defined
        """
        func_name = "derive_" + key
        return getattr(self, func_name, None)

    def derive_socket_path(self):
        return "{}/{}/core-agent.sock".format(
            self.config.value("core_agent_dir"),
            self.config.value("core_agent_full_name"),
        )

    def derive_core_agent_full_name(self):
        return "{name}-{version}-{triple}".format(
            name="scout_apm_core",
            version=self.config.value("core_agent_version"),
            triple=self.config.value("core_agent_triple"),
        )

    def derive_core_agent_triple(self):
        return PlatformDetection.get_triple()


class ScoutConfigDefaults(object):
    """
    Provides default values for important configurations
    """

    def name(self):
        return "Defaults"

    def __init__(self):
        self.defaults = {
            "app_server": "",
            "application_root": "",
            "core_agent_dir": "/tmp/scout_apm_core",
            "core_agent_download": True,
            "core_agent_launch": True,
            "core_agent_permissions": 700,
            "core_agent_version": "v1.1.8",  # can be an exact tag name, or 'latest'
            "disabled_instruments": [],
            "download_url": "https://s3-us-west-1.amazonaws.com/scout-public-downloads/apm_core_agent/release",  # noqa: E501
            "framework": "",
            "framework_version": "",
            "hostname": "",
            "key": "",
            "log_level": "info",
            "monitor": False,
            "name": "",
            "revision_sha": GitRevision().detect(),
            "scm_subdirectory": "",
        }

    def has_config(self, key):
        return key in self.defaults

    def value(self, key):
        return self.defaults[key]


# Always returns None to any key
class ScoutConfigNull(object):
    """
    Always answers that a key is present, but the value is None

    Used as the last step of the layered configuration.
    """

    def name(self):
        return "Null"

    def has_config(self, key):
        return True

    def value(self, key):
        return None


string_type = str if sys.version_info[0] >= 3 else basestring  # noqa: F821


class BooleanConversion(object):
    @classmethod
    def convert(cls, value):
        if isinstance(value, bool):
            return value
        if isinstance(value, string_type):
            return value.lower() in ("yes", "true", "t", "1")
        # Unknown type - default to false?
        return False


class ListConversion(object):
    @classmethod
    def convert(cls, value):
        if isinstance(value, list):
            return value
        if isinstance(value, tuple):
            return list(value)
        if isinstance(value, string_type):
            # Split on commas
            return [item.strip() for item in value.split(",") if item]
        # Unknown type - default to empty?
        return []


CONVERSIONS = {
    "core_agent_download": BooleanConversion,
    "core_agent_launch": BooleanConversion,
    "monitor": BooleanConversion,
    "disabled_instruments": ListConversion,
    "ignore": ListConversion,
}
